@@ -1,3 +1,4 @@
+# Copyright 2012 Urban Airship
 import collections
 import Queue as queue
 import random
@@ -7,7 +8,7 @@ import time
 
 # Pool classes/defaults
 CandidatePool = collections.deque
-ConnectionPool = queue.Queue
+ConnectionPool = queue.LifoQueue
 DEFAULT_POOL_MAX = 10
 DEFAULT_POOL_TIMEOUT = 2
 DEFAULT_CONNECT_ATTEMPTS = 2
@@ -57,7 +58,7 @@ class _ConnectionContext(object):
         # in the pool. The next get() will determine it's validity.
         self.pool.put(self.conn)
         if exc_type is not None:
-            # Let handle_excption suppress the exception if desired
+            # Let handle_exception suppress the exception if desired
             return bool(self.conn.handle_exception(exc_value))
 
 
@@ -79,8 +80,10 @@ class KiddieConnection(object):
         self.timeout = timeout
         self.last_touch = 0
         self.born = time.time()
-        if lifetime is not None:
-            self.lifetime = self.born + lifetime
+        # lifetime is how many seconds the connection lives
+        self.lifetime = lifetime
+        # endoflife is when the connection should die and be reconnected
+        self.endoflife = 0
 
     @property
     def closed(self):
@@ -104,8 +107,14 @@ class KiddieConnection(object):
         if self.tcp_keepalives:
             self.socket.setsockopt(
                     socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
+
+        # Touch to update the idle check
         self.touch()
 
+        # Reset endoflife
+        if self.lifetime is not None:
+            self.endoflife = time.time() + self.lifetime
+
     def touch(self):
         self.last_touch = time.time()
 
@@ -123,7 +132,17 @@ class KiddieConnection(object):
         return data
 
     def recvall(self, size):
-        return self.recv(size, socket.MSG_WAITALL)
+        """Receive `size` data and return it or raise a socket.error"""
+        data = []
+        received = 0
+        while received < size:
+            chunk = self.recv(size - received)
+            if not chunk:
+                raise KiddieClientRecvFailure(
+                        'Received %d bytes out of %d.' % (received, size))
+            data.append(chunk)
+            received += len(chunk)
+        return b"".join(data)
 
     def handle_exception(self, e):
         """Close connection on socket errors"""
@@ -144,7 +163,7 @@ class KiddieConnection(object):
             # Invalid because it's been idle too long
             return False
 
-        if self.lifetime is not None and self.lifetime < now:
+        if self.lifetime is not None and self.endoflife < now:
             # Invalid because it's outlived its lifetime
             return False
 
@@ -290,3 +309,35 @@ class KiddieClient(object):
             raise self.SendException(
                     'Failed to send request (%d bytes) after %d attempts. '
                     'Last exception: %r' % (len(request), attempts, e))
+
+
+class FakeConnection(object):
+    """Connection class for testing (noops)"""
+    def __init__(self, *args, **kwargs):
+        self.host = kwargs.get('host', 'localhost')
+        self.port = kwargs.get('port', 9000)
+        self.closed = kwargs.get('closed', False)
+
+    def connect(self, host, port):
+        self.host = host
+        self.port = port
+        return True
+
+    def close(self):
+        self.closed = True
+
+    def sendall(self, payload):
+        pass
+
+    def recv(self, size, flags=0):
+        return ''
+
+    def recvall(self, size):
+        return ''
+
+    def handle_exception(self, e):
+        if isinstance(e, socket.error):
+            self.close()
+
+    def validate(self):
+        return True
